@@ -25,9 +25,9 @@
 #include <errno.h>
 #include <sys/types.h>
 #include <dirent.h>
-#include <libgnomevfs/gnome-vfs.h>
 
 #include "menu-util.h"
+#include "menu-monitor.h"
 #include "canonicalize.h"
 
 typedef struct CachedDir CachedDir;
@@ -58,8 +58,8 @@ struct CachedDir
   GSList *entries;
   GSList *subdirs;
 
-  GnomeVFSMonitorHandle *monitor;
-  GSList                *monitors;
+  MenuMonitor *dir_monitor;
+  GSList      *monitors;
 
   guint have_read_entries : 1;
   guint no_monitor_support : 1;
@@ -76,6 +76,11 @@ static void     cached_dir_free                   (CachedDir  *dir);
 static gboolean cached_dir_load_entries_recursive (CachedDir  *dir,
                                                    const char *dirname);
 
+static void handle_cached_dir_changed (MenuMonitor      *monitor,
+				       MenuMonitorEvent  event,
+				       const char       *path,
+				       CachedDir        *dir);
+
 /*
  * Entry directory cache
  */
@@ -120,9 +125,14 @@ cached_dir_free (CachedDir *dir)
   cached_dir_clear_entries (dir);
   cached_dir_clear_subdirs (dir);
 
-  if (dir->monitor)
-    gnome_vfs_monitor_cancel (dir->monitor);
-  dir->monitor = NULL;
+  if (dir->dir_monitor)
+    {
+      menu_monitor_remove_notify (dir->dir_monitor,
+				  (MenuMonitorNotifyFunc) handle_cached_dir_changed,
+				  dir);
+      menu_monitor_unref (dir->dir_monitor);
+      dir->dir_monitor = NULL;
+    }
 
   g_slist_foreach (dir->monitors, (GFunc) g_free, NULL);
   g_slist_free (dir->monitors);
@@ -347,7 +357,6 @@ cached_dir_update_entry (CachedDir  *dir,
         {
           if (!desktop_entry_reload (tmp->data))
 	    {
-	      desktop_entry_unref (tmp->data);
 	      dir->entries = g_slist_delete_link (dir->entries, tmp);
 	    }
 
@@ -357,7 +366,7 @@ cached_dir_update_entry (CachedDir  *dir,
       tmp = tmp->next;
     }
 
-  return FALSE;
+  return cached_dir_add_entry (dir, basename, path);
 }
 
 static gboolean
@@ -411,7 +420,7 @@ cached_dir_remove_subdir (CachedDir  *dir,
 {
   GSList *tmp;
 
-  tmp = dir->entries;
+  tmp = dir->subdirs;
   while (tmp != NULL)
     {
       CachedDir *subdir = tmp->data;
@@ -430,29 +439,40 @@ cached_dir_remove_subdir (CachedDir  *dir,
 }
 
 static void
-handle_cached_dir_changed (GnomeVFSMonitorHandle    *handle,
-                           const char               *monitor_uri,
-                           const char               *info_uri,
-                           GnomeVFSMonitorEventType  event,
-                           CachedDir                *dir)
+cached_dir_invoke_monitors (CachedDir *dir)
+{
+  GSList *tmp;
+
+  tmp = dir->monitors;
+  while (tmp != NULL)
+    {
+      CachedDirMonitor *monitor = tmp->data;
+      GSList           *next    = tmp->next;
+
+      monitor->callback (monitor->ed, monitor->user_data);
+
+      tmp = next;
+    }
+
+  if (dir->parent)
+    {
+      cached_dir_invoke_monitors (dir->parent);
+    }
+}
+
+static void
+handle_cached_dir_changed (MenuMonitor      *monitor,
+			   MenuMonitorEvent  event,
+			   const char       *path,
+                           CachedDir        *dir)
 {
   gboolean  handled = FALSE;
-  char     *path;
   char     *basename;
 
-  if (event != GNOME_VFS_MONITOR_EVENT_CREATED &&
-      event != GNOME_VFS_MONITOR_EVENT_DELETED &&
-      event != GNOME_VFS_MONITOR_EVENT_CHANGED)
-    return;
-
-  if (!(path = gnome_vfs_get_local_path_from_uri (info_uri)))
-    return;
-
   menu_verbose ("Notified of '%s' %s - invalidating cache\n",
                 path,
-                event == GNOME_VFS_MONITOR_EVENT_CREATED ? ("created") :
-                event == GNOME_VFS_MONITOR_EVENT_DELETED ? ("deleted") :
-                event == GNOME_VFS_MONITOR_EVENT_CHANGED ? ("changed") : ("unknown-event"));
+                event == MENU_MONITOR_EVENT_CREATED ? ("created") :
+                event == MENU_MONITOR_EVENT_DELETED ? ("deleted") : ("changed"));
 
   basename = g_path_get_basename (path);
 
@@ -461,15 +481,15 @@ handle_cached_dir_changed (GnomeVFSMonitorHandle    *handle,
     {
       switch (event)
         {
-        case GNOME_VFS_MONITOR_EVENT_CREATED:
+        case MENU_MONITOR_EVENT_CREATED:
           handled = cached_dir_add_entry (dir, basename, path);
           break;
 
-        case GNOME_VFS_MONITOR_EVENT_CHANGED:
+        case MENU_MONITOR_EVENT_CHANGED:
           handled = cached_dir_update_entry (dir, basename, path);
           break;
 
-        case GNOME_VFS_MONITOR_EVENT_DELETED:
+        case MENU_MONITOR_EVENT_DELETED:
           handled = cached_dir_remove_entry (dir, basename);
           break;
 
@@ -482,14 +502,14 @@ handle_cached_dir_changed (GnomeVFSMonitorHandle    *handle,
     {
       switch (event)
         {
-        case GNOME_VFS_MONITOR_EVENT_CREATED:
+        case MENU_MONITOR_EVENT_CREATED:
           handled = cached_dir_add_subdir (dir, basename, path);
           break;
 
-        case GNOME_VFS_MONITOR_EVENT_CHANGED:
+        case MENU_MONITOR_EVENT_CHANGED:
           break;
 
-        case GNOME_VFS_MONITOR_EVENT_DELETED:
+        case MENU_MONITOR_EVENT_DELETED:
           handled = cached_dir_remove_subdir (dir, basename);
           break;
 
@@ -500,22 +520,10 @@ handle_cached_dir_changed (GnomeVFSMonitorHandle    *handle,
     }
 
   g_free (basename);
-  g_free (path);
 
   if (handled)
     {
-      GSList *tmp;
-
-      tmp = dir->monitors;
-      while (tmp != NULL)
-        {
-          CachedDirMonitor *monitor = tmp->data;
-	  GSList           *next    = tmp->next;
-
-          monitor->callback (monitor->ed, monitor->user_data);
-
-          tmp = next;
-        }
+      cached_dir_invoke_monitors (dir);
     }
 }
 
@@ -523,23 +531,13 @@ static void
 cached_dir_ensure_monitor (CachedDir  *dir,
                            const char *dirname)
 {
-  char *uri;
-
-  if (dir->monitor != NULL || dir->no_monitor_support)
-    return;
-
-  uri = gnome_vfs_get_uri_from_local_path (dirname);
-
-  if (gnome_vfs_monitor_add (&dir->monitor,
-                             uri,
-                             GNOME_VFS_MONITOR_DIRECTORY,
-                             (GnomeVFSMonitorCallback) handle_cached_dir_changed,
-                             dir) != GNOME_VFS_OK)
+  if (dir->dir_monitor == NULL)
     {
-      dir->no_monitor_support = TRUE;
+      dir->dir_monitor = menu_get_directory_monitor (dirname);
+      menu_monitor_add_notify (dir->dir_monitor,
+			       (MenuMonitorNotifyFunc) handle_cached_dir_changed,
+			       dir);
     }
-
-  g_free (uri);
 }
 
 static gboolean
@@ -864,6 +862,11 @@ entry_directory_get_desktop (EntryDirectory *ed,
         return NULL;
 
       unprefixed_id = file_id + strlen (ed->legacy_prefix);
+
+      if (unprefixed_id[0] != '-')
+	return NULL;
+
+      unprefixed_id++;
     }
 
   entry = cached_dir_find_file_id (ed->dir,
@@ -1272,107 +1275,6 @@ entry_directory_list_get_all_desktops (EntryDirectoryList *list,
   entry_directory_list_add (list, get_all_func, set, NULL);
 }
 
-static gboolean
-get_by_category_func (EntryDirectory  *ed,
-                      DesktopEntry    *entry,
-                      const char      *relative_path,
-                      const char      *file_id,
-                      DesktopEntrySet *set,
-                      gpointer         user_data)
-{
-  const char *category = user_data;
-
-  if (ed->is_legacy && !desktop_entry_has_categories (entry))
-    {
-      if (strcmp (category, "Legacy") != 0)
-        return TRUE;
-
-      entry = desktop_entry_copy (entry);
-      desktop_entry_add_legacy_category (entry);
-    }
-  else
-    {
-      if (!desktop_entry_has_category (entry, category))
-        return TRUE;
-
-      entry = desktop_entry_ref (entry);
-    }
-
-  desktop_entry_set_add_entry (set,
-                               entry,
-                               file_id ? file_id : relative_path);
-  desktop_entry_unref (entry);
-
-  return TRUE;
-}
-
-void
-entry_directory_list_get_by_category (EntryDirectoryList *list,
-                                      const char         *category,
-                                      DesktopEntrySet    *set)
-{
-  menu_verbose (" Storing list %p category %s in set %p\n",
-                list, category, set);
-
-  entry_directory_list_add (list,
-                            get_by_category_func,
-                            set,
-                            (char *) category);
-}
-
-static gboolean
-get_inverse_func (EntryDirectory  *ed,
-                  DesktopEntry    *entry,
-                  const char      *relative_path,
-                  const char      *file_id,
-                  DesktopEntrySet *set,
-                  gpointer         user_data)
-{
-  DesktopEntrySet *inverse = user_data;
-
-  /* if not in the original set, add to inverse set */
-
-  if (desktop_entry_set_lookup (set, relative_path) != NULL)
-    return TRUE;
-
-  if (ed->is_legacy && !desktop_entry_has_categories (entry))
-    {
-      entry = desktop_entry_copy (entry);
-      desktop_entry_add_legacy_category (entry);
-    }
-  else
-    {
-      entry = desktop_entry_ref (entry);
-    }
-
-  desktop_entry_set_add_entry (inverse,
-                               entry,
-                               file_id ? file_id : relative_path);
-  desktop_entry_unref (entry);
-
-  return TRUE;
-}
-
-void
-entry_directory_list_invert_set (EntryDirectoryList *list,
-                                 DesktopEntrySet    *set)
-{
-  DesktopEntrySet *inverse;
-
-  menu_verbose (" Inverting set %p relative to list %p\n",
-                set, list);
-
-  inverse = desktop_entry_set_new ();
-  entry_directory_list_add (list,
-                            get_inverse_func,
-                            set,
-                            inverse);
-
-  desktop_entry_set_swap_contents (set, inverse);
-
-  desktop_entry_set_unref (inverse);
-}
-
 void
 entry_directory_list_add_monitors (EntryDirectoryList        *list,
                                    EntryDirectoryChangedFunc  callback,
